@@ -1,4 +1,6 @@
 # -*- coding: utf-8 -*-
+# frozen_string_literal: true
+
 require 'test_helper'
 
 class EngineTest < Haml::TestCase
@@ -78,10 +80,10 @@ class EngineTest < Haml::TestCase
     "foo\n:plain\n  1\n  2\n  \#{raise 'foo'}"             => ["foo", 5],
     "= raise 'foo'\nfoo\nbar\nbaz\nbang"                   => ["foo", 1],
     "- case 1\n\n- when 1\n  - raise 'foo'"                => ["foo", 4],
-  }
+  }.freeze
 
   User = Struct.new('User', :id)
-  class CustomHamlClass < Struct.new(:id)
+  CustomHamlClass = Struct.new(:id) do
     def haml_object_ref
       "my_thing"
     end
@@ -1333,7 +1335,7 @@ HAML
   end
 
   def test_render_should_accept_a_binding_as_scope
-    string = "This is a string!"
+    string = "This is a string!".dup
     string.instance_variable_set(:@var, "Instance variable")
     b = string.instance_eval do
       var = "Local variable"
@@ -1348,18 +1350,18 @@ HAML
   end
 
   def test_yield_should_work_with_binding
-    assert_equal("12\nFOO\n", render("= yield\n= upcase", :scope => "foo".instance_eval{binding}) { 12 })
+    assert_equal("12\nFOO\n", render("= yield\n= upcase", :scope => "foo".dup.instance_eval{binding}) { 12 })
   end
 
   def test_yield_should_work_with_def_method
-    s = "foo"
+    s = "foo".dup
     engine("= yield\n= upcase").def_method(s, :render)
     assert_equal("12\nFOO\n", s.render { 12 })
   end
 
   def test_def_method_with_module
     engine("= yield\n= upcase").def_method(String, :render_haml)
-    assert_equal("12\nFOO\n", "foo".render_haml { 12 })
+    assert_equal("12\nFOO\n", "foo".dup.render_haml { 12 })
   end
 
   def test_def_method_locals
@@ -1374,7 +1376,7 @@ HAML
   end
 
   def test_render_proc_with_binding
-    assert_equal("FOO\n", engine("= upcase").render_proc("foo".instance_eval{binding}).call)
+    assert_equal("FOO\n", engine("= upcase").render_proc("foo".dup.instance_eval{binding}).call)
   end
 
   def test_haml_buffer_gets_reset_even_with_exception
@@ -2026,7 +2028,7 @@ HAML
   end
 
   def test_fake_ascii_encoding
-    assert_encoded_equal(<<HTML.force_encoding("ascii-8bit"), render(<<HAML, :encoding => "ascii-8bit"))
+    assert_encoded_equal(<<HTML.dup.force_encoding("ascii-8bit"), render(<<HAML, :encoding => "ascii-8bit"))
 <p>bâr</p>
 <p>föö</p>
 HTML
@@ -2052,7 +2054,7 @@ HAML
   end
 
   def test_encoding_error
-    render("foo\nbar\nb\xFEaz".force_encoding("utf-8"))
+    render("foo\nbar\nb\xFEaz".dup.force_encoding("utf-8"))
     assert(false, "Expected exception")
   rescue Haml::Error => e
     assert_equal(3, e.line)
@@ -2061,7 +2063,7 @@ HAML
 
   def test_ascii_incompatible_encoding_error
     template = "foo\nbar\nb_z".encode("utf-16le")
-    template[9] = "\xFE".force_encoding("utf-16le")
+    template[9] = "\xFE".dup.force_encoding("utf-16le")
     render(template)
     assert(false, "Expected exception")
   rescue Haml::Error => e
@@ -2108,15 +2110,13 @@ HAML
   end
 
   def test_block_spacing
-    begin
-      assert render(<<-HAML)
+    assert render(<<-HAML)
 - foo = ["bar", "baz", "kni"]
 - foo.each do | item |
   = item
 HAML
-    rescue ::SyntaxError
-      flunk("Should not have raised syntax error")
-    end
+  rescue ::SyntaxError
+    flunk("Should not have raised syntax error")
   end
 
   def test_tracing
@@ -2141,6 +2141,14 @@ HAML
     end
   end
 
+  def test_engine_reflects_defaults
+    defaults = Haml::Options.defaults.dup
+    Haml::Options.defaults[:attr_wrapper] = '"'
+    assert_equal %Q[<a a="a"></a>\n], Haml::Engine.new('%a(a="a")').render
+  ensure
+    Haml::Options.defaults.replace(defaults)
+  end
+
   private
 
   def assert_valid_encoding_comment(comment)
